/***************************************************************************
 *   Copyright (C) 2005 by Sergio Pistone                                  *
 *   sergio_pistone@yahoo.com.ar                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "transkodeworker.h"
#include "transkodefiletyperesolver.h"

#include <kurl.h>
#include <klocale.h>

#include <qfileinfo.h>
#include <qdir.h>
#include <qregexp.h>

// EncodeTypeFileTypeResolver static member definition
EncodeType::TEncodeType EncodeTypeFileTypeResolver::encodeType;

// Will create and register a file type resolver for use with TagLib
EncodeTypeFileTypeResolver fileTypeResolver( true );
//EncodeTypeFileTypeResolver* fileTypeResolver = new EncodeTypeFileTypeResolver( true );

//NOTE: TagLib documentation doesn't say who takes cares of deleting the added
// file type resolver. ATM, WE ARE DOING IT (since it's created on the stack).

int TransKodeJob::s_idCount = 0;
QMutex TransKodeJob::s_idCountMutex;

int TransKodeJob::getUnivocalID()
{
	s_idCountMutex.lock();

	int id = s_idCount++;

	s_idCountMutex.unlock();

	return id;
}

TransKodeJob::TransKodeJob( TransKodeWorker& w, const QString& profile, const QString& srcURL ):
	m_jobID( getUnivocalID() ),
	m_worker( w ),
	m_cancelledMutex( true ), // recursive mutex
	m_cancelled( false ),
	m_profile( String::deepCopy( profile ) ),
	m_profileOptions( 0 )
{
	m_srcURL = String::deepCopy( KURL::decode_string( srcURL ) );
	if ( m_srcURL.left( 7 ) == "file://" )
		m_srcURL = "file://" + QDir::cleanDirPath( m_srcURL.mid( 7 ) );

	readTagsFromFile( m_srcURL.mid( 8 ) );
	for ( int idx = 0; idx < Tag::SIZE; ++idx )
		m_userTags[idx] = "";
}

TransKodeJob::TransKodeJob( TransKodeWorker& w, const QString& profile, const QString& srcURL, const QString& dstURL, const QString tags[Tag::SIZE] ):
	m_jobID( getUnivocalID() ),
	m_worker( w ),
	m_cancelledMutex( true ), // recursive mutex
	m_cancelled( false ),
	m_profile( String::deepCopy( profile ) ),
	m_profileOptions( 0 )
{
	m_srcURL = String::deepCopy( KURL::decode_string( srcURL ) );
	if ( m_srcURL.left( 7 ) == "file://" )
		m_srcURL = "file://" + QDir::cleanDirPath( m_srcURL.mid( 7 ) );

	readTagsFromFile( m_srcURL.mid( 8 ) );
	for ( int idx = 0; idx < Tag::SIZE; ++idx )
		m_userTags[idx] = String::deepCopy( tags[idx] );

	m_dstURL = String::deepCopy( dstURL );
}


TransKodeJob::~TransKodeJob()
{
	if ( m_profileOptions )
		delete m_profileOptions;
}

int TransKodeJob::getID() const
{
	return m_jobID;
}

QString TransKodeJob::resolveTagForPath( Tag::TTag tag, bool quote )
{
	QString ret = String::evaluateCase(
		m_tags[tag] == "" ?
			m_worker.m_config.tagDefault( tag ) :
			m_tags[tag],
		m_worker.m_config.tagCase( tag ) );

	return quote ? String::quoteForShell( ret ) : ret;
}

QString TransKodeJob::resolveTagForFile( Tag::TTag tag, bool quote )
{
	QString ret = String::evaluateCase(
		m_tags[tag],
		m_worker.m_config.tagCase( tag ) );

	return quote ? String::quoteForShell( ret ) : ret;
}

QString TransKodeJob::generateDstPath(
	const QString& srcPath,
	const QString& dstExt,
	bool overwrite,
	bool vfat,
	const QString& replaceSpecialChars )
{
	QFileInfo srcInfo( srcPath );
	QString srcName = srcInfo.baseName( true ).replace( "}", "\\}" ).replace( "{", "\\{" );
	QString srcExt = srcInfo.extension( false ).replace( "}", "\\}" ).replace( "{", "\\{" );
	QString srcDir = srcInfo.dirPath().replace( "}", "\\}" ).replace( "{", "\\{" );;

	QString tagsForPath[Tag::SIZE];
	for ( int idx = 0; idx < Tag::SIZE; ++idx )
	{
		tagsForPath[idx] = resolveTagForPath( (Tag::TTag)idx, false );
		tagsForPath[idx].replace( "}", "\\}" ).replace( "{", "\\{" ).replace( "/", "_" );
	}

	QString paddedTrack = String::padNumber( tagsForPath[Tag::track], 2 );

	QString dstExtAux = dstExt;
	dstExtAux.replace( "}", "\\}" ).replace( "{", "\\{" ).replace( "/", "_" );

	QString dstPath = m_profileOptions->namingScheme();

	dstPath.replace( "%{src_dir}",	srcDir );
	dstPath.replace( "%{src_name}",	srcName );
	dstPath.replace( "%{src_ext}",	srcExt );
	dstPath.replace( "%{dst_ext}",	dstExtAux );

	dstPath.replace( "%{artist}",	tagsForPath[Tag::artist] );
	dstPath.replace( "%{album}",	tagsForPath[Tag::album] );
	dstPath.replace( "%{year}",		tagsForPath[Tag::year] );
	dstPath.replace( "%{track#}",	tagsForPath[Tag::track] );
	dstPath.replace( "%{track##}",	paddedTrack );
	dstPath.replace( "%{title}",	tagsForPath[Tag::title] );
	dstPath.replace( "%{comment}",	tagsForPath[Tag::comment] );
	dstPath.replace( "%{genre}",	tagsForPath[Tag::genre] );

	dstPath = String::evaluateCaseExpression( dstPath ).replace( "\\}", "}" ).replace( "\\{", "{" );

	if ( replaceSpecialChars != "" )
		for ( unsigned int idx = 0; idx < replaceSpecialChars.length() ; ++idx )
			dstPath.replace( replaceSpecialChars[idx] , "_" );

	dstPath = m_profileOptions->baseDir() + "/" + dstPath;

	if ( !overwrite )
		return System::unconflictivePath( dstPath, false, vfat );
	else
	{
		QString conflictingPath = System::alreadyExistingPath( dstPath, vfat );
		if ( conflictingPath == "" ) // there was no conflicting path
			return dstPath;
		else
		{
			QFileInfo fileInfo( conflictingPath );
			if ( fileInfo.isDir() || fileInfo.isSymLink() )
				return System::unconflictivePath( conflictingPath, true );
			else
				return conflictingPath;
		}
	}
}

bool TransKodeJob::monitorProcess( KProcess& process )
{
	bool stopped;
	do
	{
		if ( (stopped = wasCancelled()) )
		{
			process.kill( SIGKILL );
			break;
		}

		m_worker.checkPaused( process );
	}
	while ( process.isRunning() && !process.wait( 3 ) );

	return stopped;
}

int TransKodeJob::decodeToWav( const QString& srcPath, const QString& dstPath )
{
	LOG( "decodeToWav: src: " + srcPath, DECODE );
	LOG( "decodeToWav: dst: " + dstPath, DECODE );

	QString enclosedSrc = String::quoteForShell( srcPath );
	QString enclosedDst = String::quoteForShell( dstPath );

	KProcess process;
	// Things REALLY work badly when using setUseShell( true )
	// if possible should be completely avoided
	process.setPriority( m_worker.m_config.generalOpt( GeneralOpt::proc_priority ).toInt() );

	QString srcExt = QFileInfo( srcPath ).extension( false ).lower();

	if ( (srcExt == "flc" || srcExt == "flac") && m_worker.m_config.programIsSupported( Program::flac ) )
	{
		LOG( "decodeToWav: " + m_worker.m_config.programPath( Program::flac, true ) + " -d -o " + enclosedDst + " " + enclosedSrc, DECODE );
		process << m_worker.m_config.programPath( Program::flac, false ) << "-d" << "-o" << dstPath << srcPath;
	}
	else if ( (srcExt == "ape" || srcExt == "mac") && m_worker.m_config.programIsSupported( Program::mac ) )
	{
		LOG( "decodeToWav: " + m_worker.m_config.programPath( Program::mac, true ) + " " + enclosedSrc + " " + enclosedDst + " -d" , DECODE );
		process << m_worker.m_config.programPath( Program::mac, false ) << srcPath << dstPath << "-d";
	}
	else if ( srcExt == "ogg" && m_worker.m_config.programIsSupported( Program::oggdec ) )
	{
		LOG( "decodeToWav: " + m_worker.m_config.programPath( Program::oggdec, true ) + " -o " + enclosedDst + " " + enclosedSrc, DECODE );
		process << m_worker.m_config.programPath( Program::oggdec, false ) << "-o" << dstPath << srcPath;
	}
	else if ( srcExt == "mp3" && m_worker.m_config.programIsSupported( Program::lame ) )
	{
		LOG( "decodeToWav: " + m_worker.m_config.programPath( Program::lame, true ) + " --decode " + enclosedSrc + " " + enclosedDst, DECODE );
		process << m_worker.m_config.programPath( Program::lame, false ) << "--decode" << srcPath << dstPath;
	}
	else if ( srcExt == "mpc" && m_worker.m_config.programIsSupported( Program::mppdec ) )
	{
		LOG( "decodeToWav: " + m_worker.m_config.programPath( Program::mppdec, true ) + " " + enclosedSrc + " " + enclosedDst, DECODE );
		process << m_worker.m_config.programPath( Program::mppdec, false ) << srcPath << dstPath;
	}
	else if ( (srcExt == "mp4" || srcExt == "m4a" || srcExt == "aac") && m_worker.m_config.programIsSupported( Program::faad ) )
	{
		LOG( "decodeToWav: " + m_worker.m_config.programPath( Program::faad, true ) + " -o " + enclosedDst + " " + enclosedSrc, DECODE );
		process << m_worker.m_config.programPath( Program::faad, false ) << "-o" << dstPath << srcPath;
	}
	else if ( (srcExt == "spx" || srcExt == "speex") && m_worker.m_config.programIsSupported( Program::speexdec ) )
	{
		LOG( "decodeToWav: " + m_worker.m_config.programPath( Program::speexdec, true ) + " " + enclosedSrc + " " + enclosedDst, DECODE );
		process << m_worker.m_config.programPath( Program::speexdec, false ) << srcPath << dstPath;
	}
	else if ( srcExt == "wav" )
	{
		LOG( "decodeToWav: cp " + enclosedSrc + " " + enclosedDst, DECODE );
		process << "cp" << srcPath << dstPath;
	}
	else
	{
		LOG( "decodeToWav: unknown input type or primary decoder unavailable, falling back to gst-launch/mplayer", DECODE );
		process << "false";
	}

	bool stopped = false;
	int retCode = 1;
	if ( process.start( KProcess::NotifyOnExit ) )
	{
		LOG( "decodeToWav: decode started", DECODE );
		stopped = monitorProcess( process ); // blocks until the process ends
		retCode = process.exitStatus();
		process.closeAll();
		LOG( "decodeToWav: decode finished", DECODE );
	}

	if ( !stopped && retCode )
	{
		// Fallback to gst-launch if available:
		if ( m_worker.m_config.programIsSupported( Program::gstlaunch ) )
		{
			process.clearArguments();
			LOG( "decodeToWav: " + m_worker.m_config.programPath( Program::gstlaunch, true ) + " filesrc location=" + enclosedSrc + " ! spider ! wavenc ! filesink location=" + enclosedDst , DECODE );
			process << m_worker.m_config.programPath( Program::gstlaunch, false ) << "filesrc" << "location="+srcPath << "!" << "spider" << "!" << "wavenc" << "!" << "filesink" << "location="+dstPath;

			if ( process.start( KProcess::NotifyOnExit ) )
			{
				LOG( "decodeToWav: " + Program::toQString( Program::gstlaunch ) + " started", DECODE );
				stopped = monitorProcess( process ); // blocks until the process ends
				retCode = process.exitStatus();
				process.closeAll();
				LOG( "decodeToWav: " + Program::toQString( Program::gstlaunch ) + " finished", DECODE );
			}
		}

		if ( !stopped && retCode )
		{
			// NOTE: mplayer sometimes fails without giving any hint of it, so it's best to use gstreamer first if available
			process.clearArguments();
			LOG( "decodeToWav: " + m_worker.m_config.programPath( Program::mplayer, true ) + " " + enclosedSrc + " -vo null -ao pcm:file=" + enclosedDst , DECODE );
			//process << config.programPath( Program::mplayer, false ) << "-playlist -dumpstream" << enclosedSrc << "-dumpfile" << enclosedDst;
			process << m_worker.m_config.programPath( Program::mplayer, false ) << srcPath << "-vo" << "null" << "-ao" << "pcm:file=" + dstPath;

			// NOTE: mplayer requires the use of KProcess::Stdout. Otherwise it
			// will claim to have done its work but no file would have been written
			if ( process.start( KProcess::NotifyOnExit, KProcess::Stdout ) )
			{
				LOG( "decodeToWav: " + Program::toQString( Program::mplayer ) + " started", DECODE );
				stopped = monitorProcess( process ); // blocks until the process ends
				retCode = process.exitStatus();
				process.closeAll();
				LOG( "decodeToWav: " + Program::toQString( Program::mplayer ) + " finished", DECODE );
			}
		}
	}

	if ( retCode || stopped )
	{
		LOG( "decodeToWav: ERROR / exit status == " + QString::number( retCode ), retCode ? DECODE : NEVER );
		LOG( "decodeToWav: ERROR / process stopped", stopped ? DECODE : NEVER );
		return 1;
	}
	else
		return 0;
}

int TransKodeJob::encodeFromWav( const QString& srcPath, const QString& dstPath )
{
	LOG( "encodeFromWav: input:  " + srcPath, ENCODE );
	LOG( "encodeFromWav: output: " + dstPath, ENCODE );

	QString enclosedSrc = String::quoteForShell( srcPath );
	QString enclosedDst = String::quoteForShell( dstPath );

	if ( ! m_worker.m_config.encodeTypeIsSupported( m_profileOptions->encodeType() ) )
	{
		LOG( "encodeFromWav: ERROR / encoding type not supported", ENCODE );
		return 1;
	}

	KProcess process;
	// TODO: don't use setUseShell( true ) (NOTE: need a way of parsing switches for that)
	// setUseShell( true ) it's not so bad here since, in this case, there's less chance of error
	// (not supported encode types are _normally_ not processed). Still, we should get rid of it.
	process.setUseShell( true );
	process.setPriority( m_worker.m_config.generalOpt( GeneralOpt::proc_priority ).toInt() );

	switch ( m_profileOptions->encodeType() )
	{
		case EncodeType::lame:
			LOG( "encodeFromWav: " + m_worker.m_config.programPath( Program::lame, true )+ " " + m_profileOptions->switches() + " " + enclosedSrc + " " + enclosedDst, ENCODE );
			process << m_worker.m_config.programPath( Program::lame, true ) << m_profileOptions->switches() << enclosedSrc << enclosedDst; break;
		case EncodeType::mppenc:
			LOG( "encodeFromWav: " + m_worker.m_config.programPath( Program::mppenc, true )+ " " + m_profileOptions->switches() + " " + enclosedSrc + " " + enclosedDst, ENCODE );
			process << m_worker.m_config.programPath( Program::mppenc, true ) << m_profileOptions->switches() << enclosedSrc << enclosedDst; break;
		case EncodeType::oggenc:
			LOG( "encodeFromWav: " + m_worker.m_config.programPath( Program::oggenc, true )+ " " + m_profileOptions->switches() + " -o " + enclosedDst + " " + enclosedSrc, ENCODE );
			process << m_worker.m_config.programPath( Program::oggenc, true ) << m_profileOptions->switches() << "-o" << enclosedDst << enclosedSrc; break;
		case EncodeType::faac:
			LOG( "encodeFromWav: " + m_worker.m_config.programPath( Program::faac, true )+ " " + m_profileOptions->switches() + " -o " + enclosedDst + " " + enclosedSrc, ENCODE );
			process << m_worker.m_config.programPath( Program::faac, true ) << m_profileOptions->switches()
					<< "--artist" << resolveTagForFile( Tag::artist, true )
					<< "--year" << resolveTagForFile( Tag::year, true )
					<< "--album" << resolveTagForFile( Tag::album, true )
					<< "--track" << resolveTagForFile( Tag::track, true )
					<< "--title" << resolveTagForFile( Tag::title, true )
					<< "--genre" << resolveTagForFile( Tag::genre, true )
					<< "--comment" << resolveTagForFile( Tag::comment, true )
					<< "-o" << enclosedDst << enclosedSrc; break;
		case EncodeType::speexenc:
			LOG( "encodeFromWav: " + m_worker.m_config.programPath( Program::speexenc, true )+ " " + m_profileOptions->switches() + " " + enclosedSrc + " " + enclosedDst, ENCODE );
			process << m_worker.m_config.programPath( Program::speexenc, true ) << m_profileOptions->switches()
					<< "--comment artist=" + resolveTagForFile( Tag::artist, true )
					<< "--comment year=" + resolveTagForFile( Tag::year, true )
					<< "--comment album=" + resolveTagForFile( Tag::album, true )
					<< "--comment track=" + resolveTagForFile( Tag::track, true )
					<< "--title " + resolveTagForFile( Tag::title, true )
					<< "--comment comment=" + resolveTagForFile( Tag::comment, true )
					<< "--comment genre=" + resolveTagForFile( Tag::genre, true )
					<< enclosedSrc << enclosedDst; break;
		case EncodeType::flac:
			LOG( "encodeFromWav: " + m_worker.m_config.programPath( Program::flac, true )+ " " + m_profileOptions->switches() + " -F -o " + enclosedDst + " " + enclosedSrc, ENCODE );
			process << m_worker.m_config.programPath( Program::flac, true ) << m_profileOptions->switches() << "-F -o" << enclosedDst << enclosedSrc; break;
		case EncodeType::mac:
			LOG( "encodeFromWav: " + m_worker.m_config.programPath( Program::mac, true )+ " " + enclosedSrc + " " + enclosedDst + " " + m_profileOptions->switches(), ENCODE );
			process << m_worker.m_config.programPath( Program::mac, true ) << enclosedSrc << enclosedDst << m_profileOptions->switches(); break;
		case EncodeType::gstlaunch:
			LOG( "encodeFromWav: " + m_worker.m_config.programPath( Program::gstlaunch, true )+ " filesrc location= " + enclosedSrc + " ! wavparse ! " + m_profileOptions->switches() + " ! filesink location= " + enclosedDst, ENCODE );
			process << m_worker.m_config.programPath( Program::gstlaunch, true ) << "filesrc location=" << enclosedSrc << "! wavparse !" << m_profileOptions->switches() << "! filesink location=" << enclosedDst; break;
		case EncodeType::decode:
			LOG( "encodeFromWav: cp " + enclosedSrc + " " + enclosedDst, ENCODE );
			process << "cp" << enclosedSrc << enclosedDst; break;
		default:
			LOG( "encodeFromWav: ERROR / unknown encode type", ENCODE );
			return 1;
	}

	bool stopped = false;
	int retCode = 1;
	if ( process.start( KProcess::NotifyOnExit ) )
	{
		LOG( "encodeFromWav: started", ENCODE );
		stopped = monitorProcess( process );
		retCode = process.exitStatus();
		process.closeAll();
		LOG( "encodeFromWav: finished", ENCODE );
	}

	if ( retCode || stopped )
	{
		LOG( "encodeFromWav: ERROR / exit status == " + QString::number( retCode ), retCode ? ENCODE : NEVER );
		LOG( "encodeFromWav: ERROR / process stopped", stopped ? ENCODE : NEVER );
		return 1;
	}
	else
		return 0;
}

int TransKodeJob::ripStream( QStringList& paths )
{

	QString streamRipPath =	"/tmp/transkode-" + System::user() + "/" +
							QDate::currentDate().toString( "yyyy.MM.dd" ) + "_" +
							QTime::currentTime().toString( "hh:mm:ss" ) + "/";

	KProcess process;
	process.setPriority( m_worker.m_config.generalOpt( GeneralOpt::proc_priority ).toInt() );

	QString ripDuration = m_worker.m_config.generalOpt( GeneralOpt::rip_duration );

	LOG( "ripStream: "+m_worker.m_config.programPath( Program::streamripper, true )+" "+m_srcURL+" -d "+streamRipPath+" -s -o -l "+ripDuration, STREAMRIP );
	process << m_worker.m_config.programPath( Program::streamripper, false ) << m_srcURL << "-d" << streamRipPath << "-s" << "-o" << "-l" << ripDuration;

	bool stopped = false;
	int retCode = 1;
	if ( process.start( KProcess::NotifyOnExit ) )
	{
		LOG( "ripStream: started", STREAMRIP );
		stopped = monitorProcess( process );
		retCode = process.exitStatus();
		process.closeAll();
		LOG( "ripStream: finished", STREAMRIP );
	}

	if ( retCode || stopped )
	{
		LOG( "ripStream: ERROR / exit status == " + QString::number( retCode ), retCode ? STREAMRIP : NEVER );
		LOG( "ripStream: ERROR / process stopped", stopped ? STREAMRIP : NEVER );
		return 1;
	}
	else
	{
		paths = QDir( streamRipPath ).entryList ( "*", QDir::Files|QDir::Readable, QDir::Unsorted );
		for ( QStringList::Iterator it = paths.begin(); it != paths.end(); ++it )
			*it = streamRipPath + *it;

		if ( m_worker.m_config.generalOpt( GeneralOpt::rip_incomplete ) == "true" )
		{
			LOG( "ripStream: including incomplete files", STREAMRIP );
			QStringList imcompleteFiles =	QDir( streamRipPath + "incomplete/" ).
											entryList ( "*", QDir::Files|QDir::Readable, QDir::Unsorted );

			for ( QStringList::Iterator it = imcompleteFiles.begin(); it != imcompleteFiles.end(); ++it )
				paths.append( streamRipPath + "incomplete/" + *it );
		}

		for ( QStringList::Iterator it = paths.begin(); it != paths.end(); ++it )
			LOG( "ripStream: stream path: " + *it, STREAMRIP );

		if ( ! paths.size() )
			return 1;

		return 0;
	}
}

void TransKodeJob::readTagsFromFile( const QString& filePath )
{
	// Disable our fileTypeResolver for reading tags since we just guess type
	// from file extension which is what TagLib's default fileTypeResolver does
	EncodeTypeFileTypeResolver::setEncodeType( EncodeType::UNKNOWN );

	TagLib::FileRef fileRef( filePath.utf8() );
	//TagLib::FileRef fileRef( QStringToTString( filePath ) );
	if ( !fileRef.isNull() )
	{
		m_tags[Tag::artist]		= TStringToQString( fileRef.tag()->artist() );
		m_tags[Tag::album]		= TStringToQString( fileRef.tag()->album() );
		m_tags[Tag::title]		= TStringToQString( fileRef.tag()->title() );
		m_tags[Tag::genre]		= TStringToQString( fileRef.tag()->genre() );
		m_tags[Tag::comment]	= TStringToQString( fileRef.tag()->comment() );
		m_tags[Tag::year]		= QString::number( fileRef.tag()->year() );
		m_tags[Tag::track]		= QString::number( fileRef.tag()->track() );
	}
}

void TransKodeJob::writeTagsToFile( const QString& filePath )
{
	// If outputPath doesn't have an extension taglib will fail at determining the type (since, by default,
	// it's all it checks). This can be solved in some (most?) cases 'cause the file type can be determined
	// by looking at the encoding method that was used to generate the file.
	EncodeTypeFileTypeResolver::setEncodeType( m_profileOptions->encodeType() );

	TagLib::FileRef fileRef( filePath.utf8() );
	//TagLib::FileRef fileRef( QStringToTString( filePath ) );
	if ( !fileRef.isNull() )
	{
		fileRef.tag()->setArtist( QStringToTString( resolveTagForFile( Tag::artist ) ) );
		fileRef.tag()->setAlbum( QStringToTString( resolveTagForFile( Tag::album ) ) );
		fileRef.tag()->setTitle( QStringToTString( resolveTagForFile( Tag::title ) ) );
		fileRef.tag()->setGenre( QStringToTString( resolveTagForFile( Tag::genre ) ) );
		fileRef.tag()->setComment( QStringToTString( resolveTagForFile( Tag::comment ) ) );
		fileRef.tag()->setYear( m_tags[Tag::year].toInt() );
		fileRef.tag()->setTrack( m_tags[Tag::track].toInt() );
		fileRef.save();
	}
}

void TransKodeJob::transcode()
{
	LOG( "transcode: started", TRANSCODE );

	QString signalSrcURL = String::deepCopy( m_srcURL ); // NOTE: only for safe signals

	// Last chance to catch these ahead:
	m_profileOptions = m_worker.m_config.profileOptions( m_profile );
	if ( m_profileOptions == 0 ) // profile has been removed
	{
		m_worker.incProfileE();
		m_worker.emit jobProfileError( m_worker.m_workerID, m_jobID, signalSrcURL );
		m_worker.delaySignal( new JobProfileErrorEvent( m_jobID, signalSrcURL ) );
		return;
	}
	if ( !m_profileOptions->haveValidBaseDir() || !m_profileOptions->haveValidNamingScheme() ||
		 !m_worker.m_config.encodeTypeIsSupported( m_profileOptions->encodeType() ) )
	{
		m_worker.incProfileE();
		m_worker.emit jobProfileError( m_worker.m_workerID, m_jobID, signalSrcURL );
		m_worker.delaySignal( new JobProfileErrorEvent( m_jobID, signalSrcURL ) );
		return;
	}


	QStringList srcPaths;

	QRegExp regExp( "^([^:]+)://(.*)$" );
	regExp.search( m_srcURL );
	if ( regExp.cap( 1 ) != "file" )
	{
		m_worker.emit streamRipStarted( m_worker.m_workerID, m_jobID, signalSrcURL );
		m_worker.delaySignal( new StreamRipStartedEvent( m_jobID, signalSrcURL ) );

		// Check if stream rip is available
		/*if ( ! QFile::exists( m_worker.m_config.programPath( Program::streamripper ) ) )
		{

			m_worker.emit streamRipError( m_worker.m_workerID, m_jobID, signalSrcURL, i18n( "Streamripper not found" ) );
			m_worker.delaySignal( new StreamRipErrorEvent( m_jobID, signalSrcURL, i18n( "Streamripper not found" ) ) );
			return;
		}*/

		if ( ripStream( srcPaths ) )
		{
			if ( !wasCancelled() )
			{
				// if streamripper exits with error after correctly ripping some files, procced unless job was cancelled
				// TODO: el mensaje debe ser mas claro
				m_worker.emit streamRipError( m_worker.m_workerID, m_jobID, signalSrcURL, i18n( "error ripping stream" ) );
				m_worker.delaySignal( new StreamRipErrorEvent( m_jobID, signalSrcURL, i18n( "error ripping stream" ) ) );

				m_worker.m_streamRipE++;
				if ( srcPaths.size() == 0 )
					return;
			}
			else
			{
				m_worker.incJobsC();

				m_worker.emit jobCancelled( m_worker.m_workerID, m_jobID, signalSrcURL );
				m_worker.delaySignal( new JobCancelledEvent( m_jobID, signalSrcURL ) );
				return;
			}
		}
		else
		{
			m_worker.m_streamRipF++;

			QStringList rippedFiles;
			for ( QStringList::Iterator it = srcPaths.begin(); it != srcPaths.end(); ++it )
				rippedFiles.append( "file://" + *it );

			m_worker.emit streamRipFinished( m_worker.m_workerID, m_jobID, signalSrcURL, rippedFiles );
			m_worker.delaySignal( new StreamRipFinishedEvent( m_jobID, signalSrcURL, rippedFiles ) );
		}
	}
	else
		srcPaths.append( regExp.cap( 2 ) );

	for ( QStringList::Iterator it = srcPaths.begin(); it != srcPaths.end(); ++it )
	{
		bool overwrite = m_worker.m_config.generalOpt( GeneralOpt::overwrite ) == "true";
		bool vfat = m_worker.m_config.generalOpt( GeneralOpt::vfat ) == "true";

		QString srcPath = *it;
		QFileInfo srcPathInfo( srcPath );
		QString srcExt = srcPathInfo.extension( false ).lower();
		readTagsFromFile( srcPathInfo.absFilePath() ); // Read tags before generating dstPath

		QString dstPath = generateDstPath(
			srcPath,
			m_profileOptions->encodeType() == EncodeType::bypass ? srcExt : m_profileOptions->extension(),
			overwrite,
			vfat,
			m_worker.m_config.generalOpt( GeneralOpt::replace_chars ) );

		QFileInfo dstPathInfo( dstPath );
		QString newDir = dstPathInfo.dirPath();

		QString signalSrcFile = String::deepCopy( "file://" + srcPath ); // NOTE: only for safe signals
		QString signalDstFile = String::deepCopy( "file://" + dstPath ); // NOTE: only for safe signals

		if ( m_profileOptions->encodeType() == EncodeType::bypass )
		{
			// NOTE: this is not interruptible
			m_worker.emit copyStarted( m_worker.m_workerID, m_jobID, signalSrcFile, signalDstFile );
			m_worker.delaySignal( new CopyStartedEvent( m_jobID, signalSrcFile, signalDstFile ) );

			QStringList createdDirs;
			if ( !System::recursiveMakeDir( newDir, &createdDirs ) )
			{
				for ( QStringList::Iterator it = createdDirs.begin(); it != createdDirs.end(); ++it )
					QDir().rmdir( *it );

				m_worker.m_copyE++;
				m_worker.emit copyError( m_worker.m_workerID, m_jobID, signalSrcFile, i18n( "error creating target directory" ) );
				m_worker.delaySignal( new CopyErrorEvent( m_jobID, signalSrcFile, i18n( "error creating target directory" ) ) );
				continue;
			}

			if ( !System::copy( srcPath, dstPath ) )
			{
				m_worker.m_copyE++;
				m_worker.emit copyError( m_worker.m_workerID, m_jobID, signalSrcFile, i18n( "error copying file" ) );
				m_worker.delaySignal( new CopyErrorEvent( m_jobID, signalSrcFile, i18n( "error copying file" ) ) );
			}

			m_worker.m_cSrcFiles.append( signalSrcFile );
			m_worker.m_cDstFiles.append( signalDstFile );
			m_worker.emit copyFinished( m_worker.m_workerID, m_jobID, signalSrcFile, signalDstFile );
			m_worker.delaySignal( new CopyFinishedEvent( m_jobID, signalSrcFile, signalDstFile ) );
		}
		else
		{
			m_worker.emit transcodeStarted( m_worker.m_workerID, m_jobID, signalSrcFile, signalDstFile );
			m_worker.delaySignal( new TranscodeStartedEvent( m_jobID, signalSrcFile, signalDstFile ) );

			QString tempPath = m_profileOptions->baseDir() + "/.transkode_temp.wav";
			QString backupPath = m_profileOptions->baseDir() + "/.transkode_backup";

			System::remove( tempPath ); // in case the temporary file is there from before
			if ( decodeToWav( srcPath, tempPath ) )
			{
				System::remove( tempPath );

				if ( !wasCancelled() )
				{
					m_worker.m_transcodeE++;
					m_worker.emit transcodeError( m_worker.m_workerID, m_jobID, signalSrcFile, i18n( "error decoding file" ) );
					m_worker.delaySignal( new TranscodeErrorEvent( m_jobID, signalSrcFile, i18n( "error decoding file" ) ) );
					continue;
				}
				else
				{
					m_worker.incJobsC();

					m_worker.emit jobCancelled( m_worker.m_workerID, m_jobID, signalSrcURL );
					m_worker.delaySignal( new JobCancelledEvent( m_jobID, signalSrcURL ) );

					return;
				}
			}

			QStringList createdDirsList;
			if ( !System::recursiveMakeDir( newDir, &createdDirsList ) )
			{
				for ( QStringList::Iterator it = createdDirsList.begin(); it != createdDirsList.end(); ++it )
					QDir().rmdir( *it );

				m_worker.m_transcodeE++;
				m_worker.emit transcodeError( m_worker.m_workerID, m_jobID, signalSrcFile, i18n( "error creating target directory" ) );
				m_worker.delaySignal( new TranscodeErrorEvent( m_jobID, signalSrcFile, i18n( "error creating target directory" ) ) );
				continue;
			}

			QString backupOrigPath;
			bool backedUp = false;
			if ( overwrite )
			{
				LOG( "transcode: checking if outputFile already exists", TRANSCODE );
				if ( (backupOrigPath = System::alreadyExistingPath( dstPath, vfat )) != "" )
				{
					LOG( "transcode: outputFile found, backing up", TRANSCODE );
					System::move( backupOrigPath, backupPath );

					backedUp = true;
				}
				else
					LOG( "transcode: outputFile not found", TRANSCODE );
			}

			m_worker.emit transcodeProgress( m_worker.m_workerID, m_jobID, 50 );
			m_worker.delaySignal( new TranscodeProgressEvent( m_jobID, 50 ) );

			if ( encodeFromWav( tempPath, dstPath ) )
			{
				System::remove( dstPath );
				System::remove( tempPath );

				if ( backedUp ) // Restore saved file
					System::move( backupPath, backupOrigPath );
				for ( QStringList::Iterator it = createdDirsList.begin(); it != createdDirsList.end(); ++it )
					QDir().rmdir( *it );

				if ( !wasCancelled() )
				{
					m_worker.m_transcodeE++;
					m_worker.emit transcodeError( m_worker.m_workerID, m_jobID, signalSrcFile, i18n( "error encoding file" ) );
					m_worker.delaySignal( new TranscodeErrorEvent( m_jobID, signalSrcFile, i18n( "error encoding file" ) ) );
					continue;
				}
				else
				{
					m_worker.incJobsC();

					m_worker.emit jobCancelled( m_worker.m_workerID, m_jobID, signalSrcURL );
					m_worker.delaySignal( new JobCancelledEvent( m_jobID, signalSrcURL ) );

					return;
				}
			}

			if ( backedUp )
				System::remove( backupPath );
			System::remove( tempPath );

			m_worker.m_tSrcFiles.append( signalSrcFile );
			m_worker.m_tDstFiles.append( signalDstFile );
			m_worker.emit transcodeFinished( m_worker.m_workerID, m_jobID, signalSrcFile, signalDstFile );
			m_worker.delaySignal( new TranscodeFinishedEvent( m_jobID, signalSrcFile, signalDstFile ) );
		}

		writeTagsToFile( dstPath );
	}

	LOG( "transcode: finished", TRANSCODE );
}

void TransKodeJob::cancel()
{
	m_cancelledMutex.lock();

	m_cancelled = true;

	m_cancelledMutex.unlock();
}

bool TransKodeJob::wasCancelled() const
{
	bool ret;

	m_cancelledMutex.lock();

	ret = m_cancelled;

	m_cancelledMutex.unlock();

	return ret;
}
