#ifndef _TRANSKODEWORKER_H_
#define _TRANSKODEWORKER_H_

/***************************************************************************
 *   Copyright (C) 2005 by Sergio Pistone                                  *
 *   sergio_pistone@yahoo.com.ar                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <commondefs.h>

#include <transkodeconfig.h>
#include "transkodefilejob.h"
#include "transkodeevent.h"

#include <qobject.h>
#include <qthread.h>
#include <qwaitcondition.h>
#include <qmutex.h>
#include <qptrlist.h>

#include <kprocess.h>

/**
	NOTE: TransKodeWorker threads cannot be restarted as QThread threads can.
	This is due to the fact that control information (threadStopped) cannot be
	enforced to be (re)set prior to execution of run() (except in the constructor).
	It should be set DURING start() execution by the calling thread (prior to
	the actual thread creation) but start is not virtual so it cannot be made
	to include additional set up. It has to be that way so that if a thread does
		...
		worker.start();
		worker.stop();
		...
	worker is guaranteed to end at some point. Problem is that if threadStopped
	is initialized on run() (which executes on some arbitrary moment after the
	call to start) the value set by stop might get overwritten.
*/

class TransKodeWorker : public QObject, public QThread
{
	Q_OBJECT

	friend class TransKodeJob;

	public:
		TransKodeWorker( const TransKodeConfig& config );
		virtual ~TransKodeWorker();

		int getID() const;

		int  addJob( const QString& URL, const QString& profile );
		void addJobs( const QStringList& URLs, const QString& profile );

		int  addJob( const QString& srcURL, const QString& dstURL, const QString& profile, const QString tags[Tag::SIZE] );

		void removeJob( int jobID );
		void clearJobs();

		int currentJobID() const;
		bool isCurrentJob( int jobID ) const;
		int jobsQueued() const;

		void pause( unsigned long msecs = ULONG_MAX );
		void resume();
		bool paused() const;

		void end();
		bool ended() const;

	signals:

		/**
		 * Warning! this signals ARE NOT emitted from the main thread (check qt docs on threads)
		 */

		void jobProfileError( int workerID, int jobID, const QString& srcURL );
		void jobCancelled( int workerID, int jobID, const QString& srcURL );

		void streamRipStarted( int workerID, int jobID, const QString& srcURL );
		void streamRipFinished( int workerID, int jobID, const QString& srcURL, const QStringList& outputFiles );
		void streamRipError( int workerID, int jobID, const QString& srcURL, const QString& errorMsg );

		void copyStarted( int workerID, int jobID, const QString& srcFile, const QString& dstFile );
		void copyFinished( int workerID, int jobID, const QString& srcFile, const QString& dstFile );
		void copyError( int workerID, int jobID, const QString& srcFile, const QString& errorMsg );

		void transcodeStarted( int workerID, int jobID, const QString& srcFile, const QString& dstFile );
		void transcodeFinished( int workerID, int jobID, const QString& srcFile, const QString& dstFile );
		void transcodeError( int workerID, int jobID, const QString& srcFile, const QString& errorMsg );
		void transcodeProgress( int workerID, int jobID, int progress );

		void roundFinished( int workerID );

		void roundStats(
			int workerID,
			int streamRipE, int streamRipF,
			int transcodeE, int transcodeF,
			int copyE, int copyF,
			int jobsC, int profileE );

		void roundStats(
			int workerID,
			int streamRipE, int streamRipF,
			int transcodeE, const QStringList& tSrcFiles, const QStringList& tDstFiles,
			int copyE, const QStringList& cSrcFiles, const QStringList& cDstFiles,
			int jobsC, int profileE );


		/**
		 * This signals ARE emitted from the main thread
		 */

		void mtJobProfileError( int workerID, int jobID, const QString& srcURL );
		void mtJobCancelled( int workerID, int jobID, const QString& srcURL );

		void mtStreamRipStarted( int workerID, int jobID, const QString& srcURL );
		void mtStreamRipFinished( int workerID, int jobID, const QString& srcURL, const QStringList& outputFiles );
		void mtStreamRipError( int workerID, int jobID, const QString& srcURL, const QString& errorMsg );

		void mtCopyStarted( int workerID, int jobID, const QString& srcFile, const QString& dstFile );
		void mtCopyFinished( int workerID, int jobID, const QString& srcFile, const QString& dstFile );
		void mtCopyError( int workerID, int jobID, const QString& srcFile, const QString& errorMsg );

		void mtTranscodeStarted( int workerID, int jobID, const QString& srcFile, const QString& dstFile );
		void mtTranscodeFinished( int workerID, int jobID, const QString& srcFile, const QString& dstFile );
		void mtTranscodeError( int workerID, int jobID, const QString& srcFile, const QString& errorMsg );
		void mtTranscodeProgress( int workerID, int jobID, int progress );

		void mtRoundFinished( int workerID );

		void mtRoundStats(
			int workerID,
			int streamRipE, int streamRipF,
			int transcodeE, int transcodeF,
			int copyE, int copyF,
			int jobsC, int profileE );

		void mtRoundStats(
			int workerID,
			int streamRipE, int streamRipF,
			int transcodeE, const QStringList& tSrcFiles, const QStringList& tDstFiles,
			int copyE, const QStringList& cSrcFiles, const QStringList& cDstFiles,
			int jobsC, int profileE );


	protected slots:
		void slotProfilesRemoved( const QStringList& profiles );
		void slotProfilesModified( const QStringList& profiles );
		void slotEncTypesRemoved( const QValueList<int>& encodeTypes );


	protected:

		virtual void run(); // reimplemented from QThread

		TransKodeJob* getNextJob();
		void checkPaused( KProcess& process );

		void slotConfig( bool saved );

		void delaySignal( TransKodeWorkerEvent* event );
		virtual void customEvent( QCustomEvent * );

		void incProfileE();
		void resetProfileE();
		int profileE();

		void incJobsC();
		void resetJobsC();
		int jobsC();

		static int getUnivocalID();
		static int s_idCount;
		static QMutex s_idCountMutex;

		const int m_workerID;

		const TransKodeConfig& m_config;

		mutable QMutex m_threadPausedMutex;
		unsigned long m_threadPaused;
		QWaitCondition m_pauseCondition;

		typedef QPtrList<TransKodeJob> TransKodeJobList;

		mutable QMutex m_jobsListMutex;
		bool m_threadStopped;
		QWaitCondition m_jobCondition;
		TransKodeJobList m_jobsList;

		TransKodeJob* m_currentJob;
		mutable QMutex m_currentJobMutex;


		// stats variables:
		mutable QMutex m_statsMutex;
		int m_profileE;
		int m_jobsC;

		int m_streamRipE;
		int m_streamRipF;

		int m_transcodeE;
		QStringList m_tSrcFiles;
		QStringList m_tDstFiles;

		int m_copyE;
		QStringList m_cSrcFiles;
		QStringList m_cDstFiles;

	private:
 		// hide these two:
		TransKodeWorker( const TransKodeWorker& worker );
		TransKodeWorker& operator=( const TransKodeWorker& worker );
};

#endif
